# This file is part of vdr-webvideo-plugin.
#
# Copyright 2009 Antti Ajanki <antti.ajanki@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""This is a wrapper for using pycurl objects in asyncore.

Prepare a pycurl.Curl() object and create a curl_dispatcher passing
the Curl object to it. When all curl_dispatcher and
asyncore.dispatcher objects are ready, call asyncurl.loop() (instead
of asyncore.loop()).
"""

import asyncore
import logging
import pycurl

log = logging.getLogger('webvid')

def poll(timeout=0.0, map=None):
    multi_dispatcher.update_fds()
    asyncore.poll(timeout, map)
    multi_dispatcher.handle_curl_event()

def loop(timeout=30.0, use_poll=False, map=None, count=None):
    if map is None:
        map = asyncore.socket_map

    if use_poll:
        log.error('FIXME: asyncurl.loop() with use_poll==True is not implemented')
    poll_fun = poll

    if count is None:
        while map:
            poll_fun(timeout, map)
    else:
        while map and count > 0:
            poll_fun(timeout, map)
            count = count - 1


class curl_multi_dispatcher:
    """A dispatcher for pycurl.CurlMulti() objects. An instance of
    this class is created automatically. There is usually no need to
    construct one manually."""
    def __init__(self, map=None):
        if map is None:
            self._map = asyncore.socket_map
        else:
            self._map = map
        self.activefds = []
        self.completedcallbacks = {}
        self._curlm = pycurl.CurlMulti()
        self.perform_called = False
        self.deferred_delete = []
        self.has_readable_fds = False
        self.has_writable_fds = False

    def readable(self):
        return self.has_readable_fds

    def writable(self):
        return self.has_writable_fds

    def handle_read_event(self):
        # Dummy function. The actual work is done by a later call to
        # handle_curl_event()
        pass
    
    def handle_write_event(self):
        # Dummy function. The actual work is done by a later call to
        # handle_curl_event()
        pass

    def handle_expt_event(self):
        pass

    def handle_error(self):
        log.exception('Exception occurred during processing of a curl request')

    def handle_curl_event(self):
        self.perform_called = True
        self.deferred_delete = []
        while True:
            ret, num_handles = self._curlm.perform()
            if ret != pycurl.E_CALL_MULTI_PERFORM:
                break
        if ret != pycurl.E_OK:
            pass
        self.perform_called = False

        # Check if some requests have finished.
        nmsg, success, failed = self._curlm.info_read()
        for handle in success:
            try:
                func = self.completedcallbacks.get(handle, None)
                if func is not None:
                    func(0, None)
            except:
                self.handle_error()
            self.del_curl(handle)
        for handle, err, errmsg in failed:
            try:
                func = self.completedcallbacks.get(handle, None)
                if func is not None:
                    func(err, errmsg)
            except:
                self.handle_error()
            self.del_curl(handle)
        for handle in self.deferred_delete:
            self.del_curl(handle)

    def attach(self, curl, completedfunc):
        try:
            self._curlm.add_handle(curl)
        except pycurl.error:
            # the curl object is already on this multi-stack
            pass
        self.completedcallbacks[curl] = completedfunc
        self.handle_curl_event()
        self.update_fds()

    def detach(self, curl):
        if self.perform_called:
            # If this is called from a callback during a
            # CurlMulti.perform() call, wait until the call is
            # completed before deleting the curl handle.
            self.deferred_delete.append(curl)
        else:
            self.del_curl(curl)

    def del_curl(self, curl):
        try:
            self._curlm.remove_handle(curl)
        except pycurl.error:
            # the curl object is not on this multi-stack
            pass
        if self.completedcallbacks.has_key(curl):
            del self.completedcallbacks[curl]

    def update_fds(self):
        # Check if there are new fds that we are not yet aware and add
        # them to asyncore map.
        rfd, wfd, efd = self._curlm.fdset()
        self.has_readable_fds = (rfd != [])
        self.has_writable_fds = (wfd != [])
        
        newfdset = rfd+wfd+efd
        for fd in newfdset:
            if fd not in self.activefds:
                self.activefds.append(fd)
                self._map[fd] = self
        # Delete finished fds from the asyncore map
        for fd in self.activefds:
            if fd not in newfdset:
                self.activefds.remove(fd)
                if self._map.has_key(fd):
                    del self._map[fd]
    

class curl_dispatcher(asyncore.dispatcher):
    """A wrapper for using pycurl.Curl handles in asyncore."""    
    def __init__(self, curl, mdispatcher=None):
        """Starts a curl download/upload using the asyncore loop. curl
        is an initialized (setopt() calls must be done already)
        pycurl.Curl object."""
        if mdispatcher is None:
            mdispatcher = multi_dispatcher
        self.curl = curl
        mdispatcher.attach(self.curl, self.handle_completed)
        self.socket = None

    # FIXME: socket.close ???

    def log_info(self, message, type='info'):
        log.info(message)

    def add_channel(self, curlmulti=None):
        if curlmulti is None:
            curlmulti = multi_dispatcher
        curlmulti.attach(self.curl, self.handle_completed)

    def del_channel(self, curlmulti=None):
        if curlmulti is None:
            curlmulti = multi_dispatcher
        curlmulti.detach(self.curl)

    def close(self):
        self.del_channel()

    def handle_completed(err, errmsg):
        """Called when the download has finished. err is a numeric
        error code (or 0 if the download was successfull) and errmsg
        is a curl error message as a string."""
        self.close()


def test():
    class Request(curl_dispatcher):
        def __init__(self, url, outfile, i):
            self.id = i
            self.outfile = outfile
            self.curl = pycurl.Curl()
            self.curl.setopt(pycurl.URL, url)
            self.curl.setopt(pycurl.FOLLOWLOCATION, 1)
            self.curl.setopt(pycurl.WRITEFUNCTION, self.handle_write)
            curl_dispatcher.__init__(self, self.curl)

        def handle_write(self, str):
            print 'writing %d bytes (%d)' % (len(str), self.id)
            self.outfile.write(str)

        def handle_completed(self, err, errmsg):
            if err != 0:
                print 'error: %d %s' % (err, errmsg)
            else:
                print 'completed'

    r1 = Request('http://www.python.org', open('python.out', 'w'), 1)
    r2 = Request('http://en.wikipedia.org/wiki/Main_Page', open('wikipedia.out', 'w'), 2)
    loop(timeout=5)
    

pycurl.global_init(pycurl.GLOBAL_DEFAULT)
try:
    multi_dispatcher
except NameError:
    multi_dispatcher = curl_multi_dispatcher()

if __name__ == '__main__':
    test()
