# This file is part of vdr-webvideo-plugin.
#
# Copyright 2009 Antti Ajanki <antti.ajanki@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import asyncore
import asynchat
import cStringIO
import logging
import pycurl
import asyncurl
import webvidaemon
import utils

WEBVID_USER_AGENT = 'webvid/%s %s' % (webvidaemon.version, pycurl.version)
log = logging.getLogger('webvid')

try:
    import libmms
except ImportError, e:
    log.warning('Failed to load libmms: ' + str(e))

def create_downloader(url, writefunc=None, headerfunc=None, donefunc=None, headers_only=False):
    if url.startswith('mms:') or url.startswith('mmsh:'):
        try:
            libmms
        except (NameError, OSError):
            if donefunc is not None:
                donefunc(501, 'MMS scheme not supported. Install libmms.')
            return None
        return MMSDownload(url, writefunc, headerfunc, donefunc, headers_only)
    else:
        return CurlDownload(url, writefunc, headerfunc, donefunc, headers_only)

def noop_callback(str):
    pass

class CurlDownload(asyncurl.curl_dispatcher):
    def __init__(self, url, writefunc=None, headerfunc=None, donefunc=None, headers_only=False):
        self.donefunc = donefunc
        self.curl = pycurl.Curl()
        self.curl.setopt(pycurl.URL, url)
        self.curl.setopt(pycurl.FOLLOWLOCATION, 1)
        self.curl.setopt(pycurl.FAILONERROR, 1)
        self.curl.setopt(pycurl.USERAGENT, WEBVID_USER_AGENT)
        if headers_only:
            self.curl.setopt(pycurl.NOBODY, 1)
        if headerfunc is not None:
            self.curl.setopt(pycurl.HEADERFUNCTION, headerfunc)
        self.buffer = cStringIO.StringIO()
        if writefunc is None:
            self.curl.setopt(pycurl.WRITEFUNCTION, self.writetobuf)
        else:
            self.curl.setopt(pycurl.WRITEFUNCTION, writefunc)
        asyncurl.curl_dispatcher.__init__(self, self.curl)

    def writetobuf(self, msg):
        self.buffer.write(msg)

    def get_url(self):
        return self.curl.getinfo(pycurl.EFFECTIVE_URL)

    def convert_curl_error(self, err):
        """Convert a curl error code err to webvid error code."""
        # The error constants are defined only in pycurl 7.16.1 and
        # newer.
        curlver = pycurl.version_info()[2]
        if curlver >= 0x071001:
            error_code_mapping = \
                {pycurl.E_OK:                     0,
                 pycurl.E_OPERATION_TIMEOUTED:    505,
                 pycurl.E_OUT_OF_MEMORY:          500,
                 pycurl.E_PARTIAL_FILE:           504,
                 pycurl.E_READ_ERROR:             504,
                 pycurl.E_RECV_ERROR:             504,
                 pycurl.E_REMOTE_FILE_NOT_FOUND:  404,
                 pycurl.E_TOO_MANY_REDIRECTS:     404,
                 pycurl.E_UNSUPPORTED_PROTOCOL:   500,
                 pycurl.E_URL_MALFORMAT:          400,
                 pycurl.E_COULDNT_CONNECT:        506,
                 pycurl.E_COULDNT_RESOLVE_HOST:   506,
                 pycurl.E_COULDNT_RESOLVE_PROXY:  506,
                 pycurl.E_FILE_COULDNT_READ_FILE: 404,
                 pycurl.E_GOT_NOTHING:            504,
                 pycurl.E_HTTP_RETURNED_ERROR:    404,
                 pycurl.E_INTERFACE_FAILED:       506,
                 pycurl.E_LOGIN_DENIED:           403}
        else:
            error_code_mapping = {pycurl.E_OK:    0}

        if err not in error_code_mapping:
            return 500
        else:
            return error_code_mapping[err]

    def handle_completed(self, err, errmsg):
        self.buffer.seek(0)
        if self.donefunc is not None:
            self.donefunc(self.convert_curl_error(err), errmsg)
        # It seems that a reference to self.writetobuf forbids garbage
        # collection from deleting this object. unsetopt() or setting
        # the callback to None are not allowed. Is there a better way?
        self.curl.setopt(pycurl.WRITEFUNCTION, noop_callback)
        self.close()


class MMSDownload(asynchat.async_chat, asyncore.file_dispatcher):
    def __init__(self, url, writefunc=None, headerfunc=None, donefunc=None, headers_only=False):
        self.url = url
        self.writefunc = writefunc
        self.headerfunc = headerfunc
        self.donefunc = donefunc
        self.inbuffer = cStringIO.StringIO()
        self.relaylen = -1
        self.expectedlen = -1
        self.child_pid = 0
        if headers_only:
            log.debug('FIXME; headers_only mode not implemented. Will download the whole file.')

        # The mms urls are handled by a separate process
        r, w = os.pipe()
        try:
            self.child_pid = os.fork()
        except OSError:
            os.close(w)
            os.close(r)
            self.close()
            self.donefunc(501, 'Failed to fork')
            return
        if self.child_pid != 0: # parent
            os.close(w)
            self.set_terminator('\r\n\r\n')
            self.expect_header = True
            # file_dispatcher.__init__ must be called after
            # async_chat.__init__ so that it can setup self.socket
            # correctly.
            asynchat.async_chat.__init__(self)
            asyncore.file_dispatcher.__init__(self, r)
        else: # child
            # The only file handle we are only going to use is w
            os.close(r)
            for fd in xrange(3,256):
                if fd != w:
                    try:
                        os.close(fd)
                    except OSError:
                        pass

            exitcode = 0
            try:
                stream = libmms.Stream(url, 1000000)
                # Output the length in a HTTP-like header field so
                # that we can use the same callbacks as with HTTP
                # downloads.
                if utils.get_url_extension(url) == 'wmv':
                    os.write(w, 'Content-Type: video/x-ms-wmv\r\n')
                elif utils.get_url_extension(url) == 'wma':
                    os.write(w, 'Content-Type: audio/x-ms-wma\r\n')
                os.write(w, 'Content-Length: %d\r\n\r\n' % stream.length())

                log.debug('expected Content-Length: %d' % stream.length())
                
                for data in stream:
                    os.write(w, data)
            except libmms.Error, e:
                if 'stream' in locals():
                    stream.close()
                log.error('%s (url was %s)' % (e.args[0], url))
                exitcode = 1
            except OSError, (errno, strerror):
                if 'stream' in locals():
                    stream.close()
                log.error('%s (url was %s)' % (strerror, url))
                exitcode = 1
            os.close(w)
            os._exit(exitcode)

    def readable(self):
        return True

    def writable(self):
        return False

    def collect_incoming_data(self, data):
        if self.expect_header:
            self.inbuffer.write(data)
        else:
            if self.relaylen + len(data) > self.expectedlen:
                # upstream is sending us more data than indicated in
                # the header, but we already send Content-Length
                # header to downstream. We have to clip rest of the
                # data here in order not to confuse downstream.
                self.writefunc(data[:(self.expectedlen-self.relaylen)])
                self.relaylen = self.expectedlen
                self.close()
            else:
                self.relaylen += len(data)
                self.writefunc(data)

    def found_terminator(self):
        if self.expect_header:
            for header in self.inbuffer.getvalue().split('\r\n'):
                self.headerfunc(header)
                fields = header.split(':', 1)
                if (len(fields) == 2) and \
                       (fields[0].lower() == 'content-length'):
                    self.expectedlen = int(fields[1])
            self.inbuffer.close()
            self.set_terminator(None)
            self.expect_header = False
            self.relaylen = 0
        else:
            self.close()
            self.donefunc(501, 'Found_terminator called in body')

    def handle_close(self):
        asynchat.async_chat.handle_close(self)

        exitcode = 0
        if self.child_pid > 0:
            try:
                pid, status = os.waitpid(self.child_pid, 0)
                exitcode = status >> 8
            except OSError:
                pass
        
        if self.relaylen < self.expectedlen:
            # We got fewer bytes than expected. Maybe the connection
            # was lost?
            log.error('relaylen < expectedlen in MMSDownload.handle_close()')
            self.donefunc(504, 'Got fewer bytes than expected')
        elif exitcode == 0:
            self.donefunc(0, '')
        else:
            self.donefunc(501, 'Child returned an error')
