# This file is part of vdr-webvideo-plugin.
#
# Copyright 2009 Antti Ajanki <antti.ajanki@iki.fi>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

version = '1.0.4'

import socket
import asyncore
import asynchat
import cStringIO
import time
import sys
import os
import os.path
import urlparse
import logging
import sys
import libxml2
import libxslt
import pycurl
import download
import utils

# TODO
#
# timeouts
# persistent connections

class WVError(Exception):
    pass

def urljoin_query_fix(base, url, allow_fragments=True):
    """urlparse.urljoin in Python 2.5 (2.6?) and older is broken in
    case url is a pure query. See http://bugs.python.org/issue1432.
    This handles correctly the case where base is a full (http) url
    and url is a query, and calls urljoin() for other cases."""
    if url.startswith('?'):
        bscheme, bnetloc, bpath, bparams, bquery, bfragment = \
            urlparse.urlparse(base, '', allow_fragments)
        bquery = url[1:]
        return urlparse.urlunparse((bscheme, bnetloc, bpath, bparams, bquery, bfragment))
    else:
        return urlparse.urljoin(base, url, allow_fragments)

def getContentUnicode(node):
    # node.getContent() returns an UTF-8 encoded sequence of bytes (a
    # string). Convert it to a unicode object.
    return unicode(node.getContent(), 'UTF-8')


class WVTPHandlerBase(asynchat.async_chat):
    def __init__(self, request, client_address, server):
        asynchat.async_chat.__init__(self, request)
        self.client_address = client_address
        self.server = server
        self.close_connection = False
        self.rfile = None
        self.log = logging.getLogger('webvid')
        self.wait_next_request()

    def log(self, message):
        self.log_info(message, 'info')
        
    def log_info(self, message, type='info'):
        message = '[%s] %s' % (self.tokenstr(), message)
        if type == 'debug':
            self.log.debug(message)
        elif type == 'warning':
            self.log.warning(message)
        elif type == 'error':
            self.log.error(message)
        else:
            self.log.info(message)

    def wait_next_request(self):
        """Get ready to handle a new request."""
        self.set_terminator('\r\n\r\n')
        # If something has already been read during the last request
        # initialize the buffer with the previous content. Also check
        # if the old content already contains the terminator!
        if self.rfile is not None:
            oldcontent = self.rfile.read()
            self.rfile = cStringIO.StringIO()
            self.rfile.write(oldcontent)
            if oldcontent.find('\r\n\r\n') != -1:
                self.found_terminator()
        else:
            self.rfile = cStringIO.StringIO()

    def request_done(self):
        if self.close_connection:
            self.close_when_done()
        else:
            self.wait_next_request()

    def collect_incoming_data(self, data):
        self.rfile.write(data)

    def found_terminator(self):
        err, msg = self.parse_headers()
        if err != 0:
            self.send_error(err, msg)
            return
        self.set_terminator(None)
        self.handle_request()

    def parse_headers(self):
        self.rfile.seek(0)
        requestline = self.rfile.readline()
        if not requestline:
            return (400, 'Bad request')
        if requestline[-2:] == '\r\n':
            requestline = requestline[:-2]
        elif requestline[-1:] == '\n':
            requestline = requestline[:-1]

        x = requestline.split()
        if len(x) == 3:
            # parse request line
            command, ref, protocol = x
            if not protocol.startswith('WVTP/1.'):
                self.log_info('Unsupported protocol: %s' % protocol, 'error')
                return (502, 'Protocol version not supported')
            self.reference = ref
            self.method = command.upper()

            # Parse the headers
            self.headers = {}
            while True:
                headerline = self.rfile.readline()
                if headerline is None:
                    break
                if headerline[-2:] == '\r\n':
                    headerline = headerline[:-2]
                elif headerline[-1:] == '\n':
                    headerline = headerline[:-1]
                # headers end with an empty line
                if len(headerline) == 0:
                    break
                x = headerline.split(':', 1)
                if len(x) == 2:
                    header = x[0].lower()
                    value = x[1].strip()
                    self.headers[header] = value
                else:
                    self.log_info('Failed to parse header line: %s'
                                  % headerline, 'warning')
        else:
            self.log_info('Failed to parse request line: %s' % requestline,
                          'error')
            return (400, 'Bad request')
        return (0, '')

    def handle_request(self):
        funcname = 'do_' + self.method.upper()
        if not hasattr(self, funcname):
            self.send_error(501, 'Unsupported method: %s' % self.method.upper())
            return
        func = getattr(self, funcname)
        func()

    def send_error(self, code, message):
        self.log_info('Error: %d: %s' % (code, message), 'error')
        self.push('WVTP/1.0 %d %s\r\n\r\n' % (code, message))
        self.request_done()

    def tokenstr(self):
        t = hash(self)
        if t >= 0:
            return hex(t)
        else:
            t = 2*sys.maxint+2+t
            h = hex(t)
            if h[-1] == 'L':
                h = h[:-1]
            return h


class WVTPHandler(WVTPHandlerBase):
    def __init__(self, request, client_address, server):
        WVTPHandlerBase.__init__(self, request, client_address, server)
        self.dl = None

    def do_DOWNLOAD(self):
        self.log_info('DOWNLOAD %s' % self.reference, 'info')
        self.start_download(self.reference)

    def do_GET(self):
        self.log_info('GET %s' % self.reference, 'info')
        if self.reference == '/mainmenu':
            self.send_reply_success(self.server.mainmenu)
        else:
            x = self.reference.split(',', 1)
            if len(x) != 2:
                self.send_error(400, 'Bad request')
                return
            self.xsltfile, self.url = x
            if self.url == '':
                doc = self.apply_xslt('', self.url, self.xsltfile)
                if doc is None:
                    self.send_error(503, 'XSLT transformation failed')
                else:
                    self.send_reply_success(doc)
            else:
                self.dl = download.create_downloader(self.url, None,
                                                     self.callback_header,
                                                     self.callback_finished_apply_xslt)

    def do_STREAM(self):
        self.log_info('STREAM %s' % self.reference, 'info')
        if self.is_asx_playlist(self.reference):
            self.dl = download.create_downloader(self.reference, None, None,
                                                 self.callback_extract_playlist_url)
        else:
            self.check_and_send_url(self.reference)

    def do_CLOSE(self):
        self.log_info('CLOSE', 'info')
        if self.dl is not None:
            try:
                self.dl.close()
            except OSError:
                # Bad file descriptor
                pass
            self.dl = None
        self.close_connection = True
        self.request_done()

    def handle_error(self):
        self.log_info('Network error. Closing connection.', 'error')
        if self.dl is not None:
            try:
                self.dl.close()
            except OSError:
                # Bad file descriptor
                pass
            self.dl = None
        self.close()

    def wait_next_request(self):
        WVTPHandlerBase.wait_next_request(self)
        self.reply_headers_sent = False
        self.contenttype = 'application/octet-stream'
        self.contentlength = 0

    def start_download(self, url):
        """Take the necessary actions to initialize a download. If url
        is an ASX playlist, read the content url from it and start to
        download the actual content. The downloaded (content) data is
        send to self.req_write() callback and possible headers to
        self.req_header()."""
        if self.is_asx_playlist(url):
            self.dl = download.create_downloader(url, None,
                                                 self.callback_header,
                                                 self.callback_playlist_loaded)
        else:
            self.dl = download.create_downloader(url, self.callback_data,
                                                 self.callback_header,
                                                 self.callback_finished)
    
    def is_asx_playlist(self, url):
        if utils.get_url_extension(url).lower() == 'asx':
            return True
        else:
            return False

    def get_url_from_asx(self, asx, asxurl):
        """Simple ASX parser. Return the content of the first <ref>
        tag."""
        try:
            doc = libxml2.htmlReadDoc(asx, asxurl, None,
                                      libxml2.HTML_PARSE_NOERROR |
                                      libxml2.HTML_PARSE_NOWARNING |
                                      libxml2.HTML_PARSE_NONET)
        except libxml2.treeError:
            self.log_info('Can\'t parse ASX file.', 'warning')
            return None
        root = doc.getRootElement()
        ret = self._get_ref_recursive(root).strip()
        doc.freeDoc()
        return ret

    def _get_ref_recursive(self, node):
        if node.name.lower() == 'ref':
            href = node.prop('href')
            if href is not None:
                return href
        c = node.children
        while c:
            res = self._get_ref_recursive(c)
            if res is not None:
                return res
            c = c.next
        return None

    def send_url(self, url):
        body = ('<?xml version="1.0"?>'
		'<mediaurl>'
                '<url>%s</url>'
                '</mediaurl>' % libxml2.newText(url))
        self.contentlength = len(body)
        self.contenttype = 'text/xml'
        self.send_headers(200, 'OK')
        self.push(body)
        self.request_done()

    def send_headers(self, code, message):
        self.push(('WVTP/1.0 %d %s\r\n'
                   'Content-Length: %s\r\n'
                    % (code, message, self.contentlength)))
        if self.contenttype is not None:
            self.push('Content-Type: %s\r\n' % self.contenttype)
        self.push('\r\n')
        self.reply_headers_sent = True

    def send_reply_success(self, body):
	self.log_info('200 OK', 'info')
        self.contenttype = 'text/xml'
        self.contentlength = len(body)
        self.send_headers(200, 'OK')
        self.push(body)
        self.request_done()

    def callback_data(self, str):
        if not self.reply_headers_sent:
            self.send_headers(200, 'Content follows')
        self.push(str)

    def callback_header(self, str):
        x = str.split(':', 1)
        if len(x) == 2:
            headername, headerdata = x
            if headername.lower() == 'content-type':
                self.contenttype = headerdata.strip()
            elif headername.lower() == 'content-length':
                self.contentlength = headerdata.strip()

    def callback_finished(self, err, errmsg):
        self.log_info('DOWNLOAD request done (err = %d)' % err, 'info')
        if err == 0:
            if not self.reply_headers_sent:
                self.send_headers(200, 'OK')
            self.request_done()
        else:
            # Send error message only if no status has been sent
            # before. If we already started relaying body, just close
            # the connection.
            if self.reply_headers_sent:
                self.close_connection = True
                self.request_done()
            else:
                self.send_error(err, errmsg)
        self.dl = None

    def callback_playlist_loaded(self, err, errmsg):
        if err == 0:
            url = self.get_url_from_asx(self.dl.buffer.getvalue(),
                                        self.dl.get_url())
            if url is None:
                self.log_info('No ref tag in ASX file!', 'warning')
                self.send_error(404, 'No ref tag in ASX file')
            else:
                if not self.is_asx_playlist(url) and url.startswith('http:'):
                    # The protocol is really "Windows Media HTTP
                    # Streaming Protocol", not plain HTTP, even though
                    # the scheme in the ASX file is "http://". We
                    # can't do MS-WMSP but luckily most MS-WMSP
                    # servers support MMS, too.
                    self.log_info('Changing scheme from http:// to mms://', 'debug')
                    url = 'mms:' + url[5:]
                self.log_info('ASX URL = %s' % url, 'debug')
                self.start_download(url)
                return
        else:
            self.send_error(err, errmsg)
        self.dl = None

    def callback_finished_apply_xslt(self, err, errmsg):
        if err == 0:
            s = self.apply_xslt(self.dl.buffer.getvalue(), self.url,
                                self.xsltfile)
            if s is None:
                self.send_error(503, 'XSLT transformation failed')
            else:
                self.send_reply_success(s)
        else:
            self.send_error(err, errmsg)
        self.dl = None

    def callback_extract_playlist_url(self, err, errmsg):
        if err == 0:
            url = self.get_url_from_asx(self.dl.buffer.getvalue(),
                                        self.dl.get_url())
            if url:
                if self.is_asx_playlist(url):
                    self.dl = download.create_downloader(url, None, None,
                                                         self.callback_extract_playlist_url)
                    return
                else:
                    if url.startswith('http:'):
                        self.log_info('Changing scheme from http:// to mms://', 'debug')
                        url = 'mms:' + url[5:]

                    self.check_and_send_url(url)
                    return
            else:
                self.send_error(503, 'XSLT tranformation failed to produce URL')
        else:
            self.send_error(err, errmsg)
        self.dl = None

    def callback_send_url(self, err, errmsg):
        self.log_info('HEAD response: ' + str(err), 'debug')
        if err == 0:
            self.send_url(self.checking_url)
        else:
            self.send_error(err, errmsg)
        self.dl = None

    def check_and_send_url(self, url):
        """Check if the target exists (currently only for HTTP URLs)
        before relaying the URL to the client."""
        if url.startswith('http://') or url.startswith('https://'):
            self.log_info('Fetching HEAD for ' + url, 'debug')
            self.checking_url = url
            self.dl = download.create_downloader(url, None, None,
                                                 self.callback_send_url, True)
        else:
            self.send_url(url)

    def apply_xslt(self, buf, url, xsltfile):
        """Apply xslt transform from file xsltfile to the string buf.
        url is the location of buf. Returns the transformed file as a
        string, or None if the transformation couldn't be completed."""
        stylesheet = libxslt.parseStylesheetFile(os.path.join(self.server.templatebasedir, xsltfile))

        if stylesheet is None:
            self.log_info('Can\'t open stylesheet %s' % xsltfile, 'warning')
            return None
        try:
            # htmlReadDoc fails if the buffer is empty but succeeds
            # (returning an empty tree) if the buffer is a single
            # space.
            if buf == '':
                buf = ' '

            # Guess whether this is an XML or HTML document.
            if buf.startswith('<?xml'):
                doc = libxml2.readDoc(buf, url, None,
                                      libxml2.XML_PARSE_NOERROR |
                                      libxml2.XML_PARSE_NOWARNING |
                                      libxml2.XML_PARSE_NONET)
            else:
                self.log_info('Using HTML parser', 'debug')
                doc = libxml2.htmlReadDoc(buf, url, None,
                                          libxml2.HTML_PARSE_NOERROR |
                                          libxml2.HTML_PARSE_NOWARNING |
                                          libxml2.HTML_PARSE_NONET)
        except libxml2.treeError:
            stylesheet.freeStylesheet()
            self.log_info('Can\'t parse XML document', 'warning')
            return None
        resultdoc = stylesheet.applyStylesheet(doc, None)
        if resultdoc is None:
            doc.freeDoc()
            stylesheet.freeStylesheet()
            self.log_info('Can\'t apply stylesheet', 'warning')
            return None
        stylesheet.freeStylesheet()
        styleshet = None
        doc.freeDoc()
        doc = None

        # Postprocess the document:
        # Add full template path to ref and object items.
        # Resolve relative URLs
        xsltdir = os.path.dirname(xsltfile)
        root = resultdoc.getRootElement()
        if root is None:
            resultdoc.freeDoc()
            self.log_info('No root node after applying XSLT transformation', 'warning')
            return None

        # Using the default iterator (for node in resultdoc)
        # segfaults, probably because we are changing the tree with
        # setContent below. Therefore the more cumbersome double loop
        # approach.
        node2 = root.children
        while node2 is not None:
            if node2.name not in ['link', 'button']:
                node2 = node2.next
                continue

            node = node2.children
            while node is not None:
                if (node.name == 'ref') or (node.name == 'object') or (node.name == 'submission'):
                    # Empty reference URL is a special case (don't
                    # download a page), leave it empty. Otherwise
                    # concatenate the (possible relative) reference
                    # URL to the original URL.
                    refurl = getContentUnicode(node)
                    if refurl == '':
                        newurl = ''
                    else:
                        newurl = urljoin_query_fix(url, refurl)
                    
                    templatefile = node.prop('template')
                    if templatefile is None:
                        templatefull = ''
                    else:
                        templatefull = os.path.join(xsltdir, templatefile + '.xsl')

                    node.unsetProp('template')
                    node.setContent(resultdoc.encodeSpecialChars(templatefull+ ',' + newurl))
                node = node.next
            node2 = node2.next
        ret = resultdoc.serialize('UTF-8')
        resultdoc.freeDoc()
        return ret
        

class WVDServer(asyncore.dispatcher):
    def __init__(self, port, handler, templatebasedir):
        self.log = logging.getLogger('webvid')
        self.log_info('webvid %s starting on port %s' % (version, port), 'info')

        asyncore.dispatcher.__init__(self)
        self.templatebasedir = templatebasedir
        self.port = port
        self.handler = handler
        self.load_mainmenu()

        self.create_socket(socket.AF_INET, socket.SOCK_STREAM)
        self.set_reuse_addr()
        self.bind(('', port))
        self.listen(5)

    def log(self, message):
        self.log_info(message, 'info')
        
    def log_info(self, message, type='info'):
        if type == 'debug':
            self.log.debug(message)
        elif type == 'warning':
            self.log.warning(message)
        elif type == 'error':
            self.log.error(message)
        else:
            self.log.info(message)

    def load_mainmenu(self):
        """Build the XML main menu from the module description files
        in the hard drive. The result string is stored in
        self.mainmenu."""
        self.mainmenu = None
        if not os.path.isdir(self.templatebasedir):
            raise WVError, 'Service directory %s does not exist' % self.templatebasedir
            return

        # Find menu items in the service.xml files in the subdirectories
        menuitems = {}
        for f in os.listdir(self.templatebasedir):
            filename = os.path.join(self.templatebasedir, f, 'service.xml')
            try:
                doc = libxml2.parseFile(filename)
            except libxml2.parserError:
                continue

            title = None
            url = None
            template = None

            root = doc.getRootElement()
            if (root is None) or (root.name != 'service'):
                self.log_info('Failed to parse %s' % filename, 'warning')
                doc.freeDoc()
                continue
            node = root.children
            while node is not None:
                if node.name == 'title':
                    title = getContentUnicode(node)
                elif node.name == 'url':
                    url = getContentUnicode(node)
                elif node.name == 'template':
                    template = getContentUnicode(node)
                node = node.next
            doc.freeDoc()

            if (title is None) or (url is None):
                self.log_info('Failed to parse %s' % filename, 'warning')
                continue

            ref = '%s/%s.xsl,%s' % (f, template, url)
            menuitems[title.lower()] = ('<link>\n'
                                        '<label>%s</label>\n'
                                        '<ref>%s</ref>\n'
                                        '</link>\n' %
                                        (libxml2.newText(title),
                                         libxml2.newText(ref)))
        # Sort the menu items
        titles = menuitems.keys()
        titles.sort()

        # Build the menu
        self.mainmenu = ('<?xml version="1.0"?>\n'
                         '<wvmenu>\n'
                         '<title>Select video source</title>\n')
        for t in titles:
            self.mainmenu += menuitems[t]
        self.mainmenu += '</wvmenu>'

    def handle_accept(self):
        try:
            conn, addr = self.accept()
        except socket.error:
            self.log_info('accept() threw an exception', 'warning')
            return
        except TypeError:
            self.log_info('accept() threw EWOULDBLOCK', 'warning')
            return
        # creates an instance of the handler class to handle the
        # request/response on the incoming connection
        h = self.handler(conn, addr, self)
	self.log_info('[%s] New connection' % h.tokenstr(), 'info')


def xml_error_callback(ctx, str):
    # be silent
     pass


# Old versions of libxslt do not have init() method
if hasattr(libxslt, 'init'):
    libxslt.init()
libxslt.registerAllExtras()
libxml2.registerErrorHandler(xml_error_callback, 'xml')
libxslt.registerErrorHandler(xml_error_callback, 'xslt')
